#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkDerivativeImageFilter.h"
#include "itkVectorIndexSelectionCastImageFilter.h"
#include "itkAddImageFilter.h"
#include "itkMultiplyImageFilter.h"
#include <vnl/vnl_math.h>
#include <itkImageRegionIterator.h>
#include <itkDivideByConstantImageFilter.h>
#include <itkWarpVectorImageFilter.h>
#include <itkWarpImageFilter.h>
#include <itkWarpVectorImageFilter.h>
#include "string.h"
#include <itkImageToVectorImageFilter.h>
#include <tclap/CmdLine.h>

/*
 * The program implements the iterative computation of the logJacobian scalar map of a deformation field 
 * parametrized by a stationary velocity field (SVF). Given the one paramter subgroups property 
 * exp(2v)=exp(v) o exp(v), the jacobian computation is iteratively computed through the following scheme:
 *
 * Step1. Find n such that v0=inputSVF/n "small" (scaling step) 
 * Step2. Approximate logJac(v0)=Div(v0)
 * Step3. Iterate n times logJac(vi)=logJac(vi-1)|v0+logJac(vi-1)) 
 *
 * for further details see Lorenzi et al, STIA Workshop, MICCAI 2010, Beijing.
 *
 * The program requires the input SVF image path and the output logJacobian map path. 
 * It is possible to provide a Mask for the computation of scaling step (useful to control boundaries),
 * and to provide a scaling factor for the input SVF.
 */


/**
 * Structure containing the input parameters.
 */
struct Param{
    std::string  SVFImage;
    std::string  Mask;
    std::string  OutputImage;
    float ScalingFactor;
    int  NumericalScheme;
};


/**
 * Parses the command line arguments and deduces the corresponding Param structure.
 * @param  argc   number of arguments
 * @param  argv   array containing the arguments
 * @param  param  structure of parameters
 */
void parseParameters(int argc, char** argv, struct Param & param)
{

    // Program description
    std::string description = "\b\b\bDESCRIPTION\n";
    description += "Iterative Log Jacobian Map Computation";
    description += "\nAuthor : Marco Lorenzi";

    try {

        // Define the command line parser
        TCLAP::CmdLine cmd( description, ' ', "1.0", true);

        TCLAP::ValueArg<std::string>  arg_SVFImage( "i", "input-svf", "Path to the input stationary velocity field", true, "", "string", cmd );
        TCLAP::ValueArg<std::string>  arg_OutputImage( "o", "output-svf", "Path of the output LogJacobian map (default LogJacobian.mha).", false, "LogJacobian.mha", "string", cmd );
        TCLAP::ValueArg<std::string>  arg_Mask( "m", "mask", "Path to the mask (default whole image)", false, "null", "string", cmd );
        TCLAP::ValueArg<double>  arg_ScalingFactor( "s", "scaling-factor", "Scaling factor for the input velocity field (default 1)", false, 1.0, "double", cmd );
    TCLAP::ValueArg<int>  arg_NumericalScheme( "z", "numerical-scheme", "Numerical scheme for the exponential: 0 Scaling and squarings (default), 1 Forward Euler", false, 0, "int", cmd );

        // Parse the command line
        cmd.parse( argc, argv );

        // Set the parameters
        param.SVFImage                     = arg_SVFImage.getValue();
        param.OutputImage                  = arg_OutputImage.getValue();
        param.Mask                         = arg_Mask.getValue();
        param.ScalingFactor                = arg_ScalingFactor.getValue();
	param.NumericalScheme              = arg_NumericalScheme.getValue();
    }
    catch (TCLAP::ArgException &e)
    {
        std::cerr << "Error: " << e.error() << " for argument " << e.argId() << std::endl;
        throw std::runtime_error("Unable to parse the command line arguments.");
    }
}


/**
  * Prints parameters.
  * @param  SVFImage                        path to the input SVF
  * @param  OutputImage      		    path to the output log jacobian map
  * @param  Mask                 	    path to the mask
  * @param  Scaling Factor                  ScalingFactor
  * @param  Numerical Scheme                Numerical Scheme
  */
void PrintParameters( std::string SVFImage,
                      std::string OutputImage,
                      std::string Mask,
                      double      ScalingFactor,
		      bool  NumericalScheme )
{

    // Print I/O parameters
    std::cout << std::endl;
    std::cout << "I/O PARAMETERS"                  << std::endl;
    std::cout << "  Input SVF image path           : " << SVFImage      << std::endl;
    std::cout << "  Output image path              : " << OutputImage     << std::endl;
    std::cout << "  Mask image path                : " << Mask     << std::endl;
    std::cout << "  Scaling Factor                 : " << ScalingFactor << std::endl;
    std::cout << "  Numerical Scheme               : " << NumericalScheme << std::endl;

    std::cout << std::endl;
}


int main( int argc, char ** argv )
{
  
  //Parsing initial parameters
  struct Param param;
  parseParameters( argc, argv, param);

  //Definition of the type:  
  typedef itk::Vector<float,3>  VectorPixelType;
  typedef itk::Image<VectorPixelType,3> VectorImageType;	
  typedef itk::Image<float,3> ImageType;

  //Definition of Reader and Writer Classes:
  typedef itk::ImageFileReader< ImageType >  ScalarReaderType;
  typedef itk::ImageFileReader< VectorImageType >  VectorReaderType;
    typedef itk::ImageFileWriter< VectorImageType >  VectorWriterType;

  ScalarReaderType::Pointer readerMask = ScalarReaderType::New();
  VectorReaderType::Pointer reader1  = VectorReaderType::New();
  VectorWriterType::Pointer writer = VectorWriterType::New();

  float mult=param.ScalingFactor;
  try
  {
   reader1->SetFileName(param.SVFImage);
   reader1->Update();
  }
  catch 
   (TCLAP::ArgException &e)
  {
   std::cerr << "Error: " << e.what() << std::endl;
   return EXIT_FAILURE;
  }

  typedef itk::ImageRegionIterator<VectorImageType> IteratorType;  
  
  VectorImageType::Pointer Out1=VectorImageType::New();
  Out1->SetRegions(reader1->GetOutput()->GetLargestPossibleRegion() );
  Out1->SetSpacing( reader1->GetOutput()->GetSpacing() );
  Out1->SetOrigin( reader1->GetOutput()->GetOrigin() );
  Out1->SetDirection( reader1->GetOutput()->GetDirection() );
  Out1->Allocate();
  
  IteratorType out1(Out1,reader1->GetOutput()->GetLargestPossibleRegion());
 
  typedef itk::Index<3> IndexType;
  itk::Vector<float,3> Vect;  
  
  for (out1.GoToBegin();!out1.IsAtEnd();++out1 ) 
    {
      Vect.Fill(0);
      const IndexType index=out1.GetIndex();
      Vect= mult*(reader1->GetOutput()->GetPixel( index ));
      out1.Set( Vect);
    }

  //replace with
  /*  m_Multiplier->SetInput(this->GetInput());
  m_Multiplier->SetConstant(m_MultiplicativeFactor);
  m_Multiplier->Update();

  InputImageConstPointer inputPtr = m_Multiplier->GetOutput();*/

  double minpixelspacing = Out1->GetSpacing()[0];
  for (unsigned int i = 0; i<3; ++i)
  {
   if ( Out1->GetSpacing()[i] < minpixelspacing )
     minpixelspacing = Out1->GetSpacing()[i];
  }

      std::cout<<"minpixelspacing  " << minpixelspacing << std::endl;
   
  IteratorType InputIt (Out1, Out1->GetRequestedRegion());

  float norm2,maxnorm2=0;
   
  if (param.Mask!="null")
   try
    {
     readerMask->SetFileName(param.Mask);
     readerMask->Update();	
    }
   catch(TCLAP::ArgException &e)
    {
     std::cerr << "Error: " << e.what() << std::endl;
     return EXIT_FAILURE;
    }

  /* 
   *   Evaluate the maximum norm in the region of interest
   */
  
  for( InputIt.GoToBegin(); !InputIt.IsAtEnd(); ++InputIt )
  {
   if ((param.Mask!="null"))
   { if (readerMask->GetOutput()->GetPixel(InputIt.GetIndex())>0)
     {
      norm2 = InputIt.Get().GetSquaredNorm();
      if (maxnorm2<norm2) 
       maxnorm2=norm2;
     }
   }  
   else 
   {
     norm2 = InputIt.Get().GetSquaredNorm();
     if (maxnorm2<norm2) 
       maxnorm2=norm2;
    }
   }

  // Divide the norm by the minimum pixel spacing
  maxnorm2 /= vnl_math_sqr(minpixelspacing);

  std::cout<<"maxnorm2 " << maxnorm2 <<std::endl;

  /**
    *    Evaluate the number N of iterations required such that maxnorm/N<=minspacing/2
   **/
  float numiterfloat;
  unsigned int numiter=0;
   unsigned int numiterEu=0;
  int divider;

     ////////////////////////////////////

     if (param.NumericalScheme==0)
     {numiterfloat = (2.0 + 0.5 * vcl_log(maxnorm2)/vnl_math::ln2); //pkoi floor
         // take the ceil and threshold
         numiter = static_cast<unsigned int>(numiterfloat + 1.0);
         divider = static_cast<float>(1<<static_cast<unsigned int>(numiter));
        if(numiter<=1)
         {numiter=1;
         divider=2;}

 }
else
     {numiter=ceil(2*sqrt(maxnorm2));
         divider=numiter;
         if(numiter<=1)
         {numiter=1;
         divider=2;
         //numiter=2;
           //       divider=2;


         }}




////////////////////////////////
  std::cout<<"numiter: " << numiter <<std::endl;
    std::cout<<"divider: " << divider <<std::endl;


    /**
    *     Scale the input SVF (i.e. first approximation)
   **/
  typedef itk::DivideByConstantImageFilter<VectorImageType,float,VectorImageType> DividerType;
  DividerType::Pointer Divider=DividerType::New();
  Divider->SetInput(Out1);
  Divider->SetConstant(divider );
  Divider->Update();

 /* DividerType::Pointer Divider2=DividerType::New();
  Divider2->SetInput(Divider->GetOutput());
  Divider2->SetConstant(-1 );
  Divider2->Update();*/


  ///////////////////////////////////////////////////////////////////
  /////////////  Exponentiation ////////////////////////
  /////////////////////////////////////////////////////////////

  typedef itk::WarpVectorImageFilter<VectorImageType,VectorImageType,VectorImageType> WarperType;
    typedef itk::AddImageFilter<VectorImageType,VectorImageType,VectorImageType> AdderType;

  AdderType::Pointer CurrentDeformation=AdderType::New();
  WarperType::Pointer VectorWarper=WarperType::New();
  VectorWarper->SetOutputOrigin(Out1->GetOrigin());
  VectorWarper->SetOutputSpacing(Out1->GetSpacing());
  VectorWarper->SetOutputDirection(Out1->GetDirection());


   VectorImageType::Pointer V=VectorImageType::New();
   VectorImageType::Pointer VoV=VectorImageType::New();

    V=Divider->GetOutput();

std::string str;
/*
  //Exponentiation of the scaled SVF
  //
if (param.NumericalScheme==2)
{ for( int i=0; i<numiter; i++ )
  {
  VectorWarper->SetInput(V);

  if (i==0)
  {VectorWarper->SetDeformationField(V);} //uiou0
  else {VectorWarper->SetDeformationField(CurrentDeformation->GetOutput());}

  VectorWarper->GetOutput()->SetRequestedRegion(V->GetRequestedRegion() );
  VectorWarper->Update();

  VoV = VectorWarper->GetOutput();

  if (i==0)
  {CurrentDeformation->SetInput1(V);}
  else
  {CurrentDeformation->SetInput1(CurrentDeformation->GetOutput());} //ui

  CurrentDeformation->SetInput2(VoV);
  CurrentDeformation->GetOutput()->SetRequestedRegion(V->GetRequestedRegion() );
  CurrentDeformation->Update();

    }}

else
  //Forward Euler scheme.
  if (param.NumericalScheme==1)
  { for( int i=0; i<numiter; i++ )
    {
    if (i==0)
    {VectorWarper->SetInput(V);}
    else
    {VectorWarper->SetInput(CurrentDeformation->GetOutput());}

    VectorWarper->SetDeformationField(V); //uiou0
    VectorWarper->GetOutput()->SetRequestedRegion(V->GetRequestedRegion() );
    VectorWarper->Update();

    VoV = VectorWarper->GetOutput();

    if (i==0)
    {CurrentDeformation->SetInput1(V);} //u0

    CurrentDeformation->SetInput2(VoV);
    CurrentDeformation->GetOutput()->SetRequestedRegion(V->GetRequestedRegion() );
    CurrentDeformation->Update();

      }}
  else
  //Scaling and Squarings scheme
  { for( int i=0; i<numiter; i++ )
   {
    VectorWarper->SetInput(V);
    VectorWarper->SetDeformationField(V);
    VectorWarper->GetOutput()->SetRequestedRegion(V->GetRequestedRegion() );
    VectorWarper->Update();

    VoV = VectorWarper->GetOutput();

    CurrentDeformation->SetInput1(V);
    CurrentDeformation->SetInput2(VoV);
    CurrentDeformation->GetOutput()->SetRequestedRegion(V->GetRequestedRegion() );
    CurrentDeformation->Update();

    V=CurrentDeformation->GetOutput();
    V->DisconnectPipeline();
   }}

  //Writing the output
  writer->SetFileName( param.OutputImage);
  writer->SetInput( CurrentDeformation->GetOutput());
*/


  ///////////////////////////////////////////////////////////////////
  /////////////  Log-Jacobian ////////////////////////
  /////////////////////////////////////////////////////////////

////////////////////////////////////////////////
  /**
    *     Initialize the iterative computation 
   **/

 // typedef itk::WarpVectorImageFilter<VectorImageType,VectorImageType,VectorImageType> WarperType;
  WarperType::Pointer Warper=WarperType::New();
  Warper->SetOutputOrigin(Out1->GetOrigin());
  Warper->SetOutputSpacing(Out1->GetSpacing());
  Warper->SetOutputDirection(Out1->GetDirection());

 // typedef itk::AddImageFilter<VectorImageType,VectorImageType,VectorImageType> AdderType;
  AdderType::Pointer Adder=AdderType::New();  
  
  VectorImageType::Pointer Vect1=VectorImageType::New();
 Vect1=Divider->GetOutput();

  VectorImageType::Pointer warpedIm=VectorImageType::New();
   
  typedef itk::WarpImageFilter<ImageType,ImageType,VectorImageType> WarpImgType;
  WarpImgType::Pointer LJoV=WarpImgType::New();
  typedef itk::AddImageFilter<ImageType,ImageType,ImageType> AddImgFilterType;
  AddImgFilterType::Pointer CurrentLJ=AddImgFilterType::New();



  /**
    *     Initial step: compute Div(v0)~logJac(v0) 
   **/

  
  typedef itk::VectorIndexSelectionCastImageFilter< VectorImageType,ImageType> FilterType;
  FilterType::Pointer componentExtractor0 = FilterType::New();
  componentExtractor0->SetIndex( 0 );
  FilterType::Pointer componentExtractor1 = FilterType::New();
  componentExtractor1->SetIndex( 1);
  FilterType::Pointer componentExtractor2 = FilterType::New();
  componentExtractor2->SetIndex( 2 );

 componentExtractor0->SetInput( Vect1 );
 componentExtractor1->SetInput( Vect1 );
 componentExtractor2->SetInput( Vect1 );
 /* componentExtractor0->SetInput( Divider2->GetOutput() );
  componentExtractor1->SetInput( Divider2->GetOutput() );
  componentExtractor2->SetInput( Divider2->GetOutput() );*/


  // Compute partial derivatives

  typedef itk::DerivativeImageFilter<
             ImageType, ImageType > DerivativeFilterType;

  DerivativeFilterType::Pointer derivative0=DerivativeFilterType::New();
  DerivativeFilterType::Pointer derivative1=DerivativeFilterType::New();
  DerivativeFilterType::Pointer derivative2=DerivativeFilterType::New();

  derivative0->SetOrder(1);
  derivative0->SetDirection(0);

  derivative1->SetOrder(1);
  derivative1->SetDirection(1);

  derivative2->SetOrder(1);
  derivative2->SetDirection(2);

  derivative0->SetInput(componentExtractor0->GetOutput());
  derivative1->SetInput(componentExtractor1->GetOutput());
  derivative2->SetInput(componentExtractor2->GetOutput());

  derivative0->Update();
  derivative1->Update();
  derivative2->Update();
  
 AddImgFilterType::Pointer addFilter = AddImgFilterType::New();
  addFilter->SetInput1( derivative0->GetOutput() );
  addFilter->SetInput2( derivative1->GetOutput() );
  addFilter->Update();

  AddImgFilterType::Pointer addFilter1 = AddImgFilterType::New();
  addFilter1->SetInput1( derivative2->GetOutput() );
  addFilter1->SetInput2(addFilter->GetOutput());
  addFilter1->Update();

  ImageType::Pointer Image=addFilter1->GetOutput();

 // typedef itk::WarpVectorImageFilter<VectorImageType,VectorImageType,VectorImageType> WarperType;
  //typedef itk::AddImageFilter<VectorImageType,VectorImageType,VectorImageType> AdderType;

  AdderType::Pointer VectorAdder=AdderType::New();  
  WarperType::Pointer VectorWarper2=WarperType::New();
  VectorWarper2->SetOutputOrigin(Out1->GetOrigin());
  VectorWarper2->SetOutputSpacing(Out1->GetSpacing());
  VectorWarper2->SetOutputDirection(Out1->GetDirection());

  WarpImgType::Pointer ImageWarper=WarpImgType::New();
  ImageWarper->SetOutputOrigin(Out1->GetOrigin());
  ImageWarper->SetOutputSpacing(Out1->GetSpacing());
  ImageWarper->SetOutputDirection(Out1->GetDirection());
  

AdderType::Pointer CurrentDeformation2=AdderType::New();
 /**
    *     Iterative step: compute logJac(vi)=logJac(vi-1)|exp(v0)+logJac(vi-1) 
   **/

std::cout<<"numiter 1 : " << numiter <<std::endl;
  std::cout<<"divider 1 : " << divider <<std::endl;

//////////////def

  VectorImageType::Pointer Vec_1=VectorImageType::New();

  if (param.NumericalScheme==2)
  { for( int i=0; i<numiter; i++ )
    {
          // mine LJ
          ImageWarper->SetInput(Image);

        if (i==0)
          {ImageWarper->SetDeformationField(Vect1);}
          else
          {ImageWarper->SetDeformationField(Vec_1);}

          ImageWarper->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
          ImageWarper->Update();

          CurrentLJ->SetInput1(ImageWarper->GetOutput()); //u0

          if (i==0)
          {CurrentLJ->SetInput2(Image);}
          else
          {CurrentLJ->SetInput2(CurrentLJ->GetOutput());}
          CurrentLJ->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
          CurrentLJ->Update();

         //////////////     def
          VectorWarper2->SetInput(Vect1);

          if (i==0)
          {VectorWarper2->SetDeformationField(Vect1);} //uiou0
          else {VectorWarper2->SetDeformationField(CurrentDeformation2->GetOutput());}

          VectorWarper2->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
          VectorWarper2->Update();

          VoV = VectorWarper2->GetOutput();

          if (i==0)
          {CurrentDeformation2->SetInput1(Vect1);}
          else
          {CurrentDeformation2->SetInput1(Vec_1);} //ui //tester si disconnect le current...

          CurrentDeformation2->SetInput2(VoV);
          CurrentDeformation2->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
          CurrentDeformation2->Update();

          Vec_1=CurrentDeformation2->GetOutput();
              Vec_1->DisconnectPipeline();
      }}


  /////////////////////////////

if (param.NumericalScheme==1)
{ for( int i=0; i<numiter; i++ )
  {
        // mine LJ
      if (i==0)
        {ImageWarper->SetInput(Image);}
        else
        {ImageWarper->SetInput(CurrentLJ->GetOutput());}

        ImageWarper->SetDeformationField(Vect1); //LJou0
        ImageWarper->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
        ImageWarper->Update();

        if (i==0)
        {CurrentLJ->SetInput1(Image);} //u0

        CurrentLJ->SetInput2(ImageWarper->GetOutput());
        CurrentLJ->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
        CurrentLJ->Update();

   //////////////     def
  if (i==0)
  {VectorWarper2->SetInput(Vect1);}
  else
  {VectorWarper2->SetInput(CurrentDeformation2->GetOutput());}

  VectorWarper2->SetDeformationField(Vect1); //uiou0
  VectorWarper2->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
  VectorWarper2->Update();

  VoV = VectorWarper2->GetOutput();

  if (i==0)
  {CurrentDeformation2->SetInput1(Vect1);} //u0

  CurrentDeformation2->SetInput2(VoV);
  CurrentDeformation2->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
  CurrentDeformation2->Update();

    }}
///////////////////////////////////////////////////////////////////////

std::cout<<"numiter: " << numiter <<std::endl;
  std::cout<<"divider: " << divider <<std::endl;

if (param.NumericalScheme==0)
  { for( int i=0; i<numiter; i++ )
    {
     LJoV->SetInput(Image);
     LJoV->SetDeformationField(Vect1);
     LJoV->SetOutputOrigin(Vect1->GetOrigin());
     LJoV->SetOutputSpacing(Vect1->GetSpacing());
     LJoV->SetOutputDirection(Vect1->GetDirection());
     LJoV->UpdateLargestPossibleRegion();
     //LJoV->Update();

     CurrentLJ->SetInput1(LJoV->GetOutput());
     CurrentLJ->SetInput2(Image);
     CurrentLJ->Update();

     //calcul de phi

     VectorWarper2->SetInput(Vect1);
     VectorWarper2->SetDeformationField(Vect1);
     VectorWarper2->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion());
     VectorWarper2->Update();


     CurrentDeformation2->SetInput1(VectorWarper2->GetOutput());
     CurrentDeformation2->SetInput2(Vect1);
     CurrentDeformation2->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
     CurrentDeformation2->Update();

     //Vect1=VectorWarper2->GetOutput();//faux
    Vect1=CurrentDeformation2->GetOutput();
    Image=CurrentLJ->GetOutput();
     Vect1->DisconnectPipeline();
     //Image->DisconnectPipeline();



   // VectorImageType::Pointer Vec_1=VectorWarper2->GetOutput();
    //Vec_1->DisconnectPipeline();
   //ImageType::Pointer Image_1=LJoV->GetOutput();
    // Image_1->DisconnectPipeline();

    }}
 // }


     VectorWriterType::Pointer writer2 = VectorWriterType::New();

     //Writing the output
     writer2->SetFileName( "Phi_" + param.OutputImage);
     //writer2->SetInput( VectorWarper2->GetOutput());
     writer2->SetInput( CurrentDeformation2->GetOutput());

  typedef itk::ImageFileWriter<ImageType> ImageWriterType;
  ImageWriterType::Pointer WriterImg=ImageWriterType::New();

  WriterImg->SetInput(CurrentLJ->GetOutput());
  WriterImg->SetFileName("LJ_" + param.OutputImage);




  try
    {
    //writer->Update();
    writer2->Update();
   WriterImg->Update();
    }
  catch( itk::ExceptionObject & err )
    {
    std::cerr << "ExceptionObject caught !" << std::endl;
    std::cerr << err << std::endl;
    return EXIT_FAILURE;}


  return EXIT_SUCCESS;

}

/*
/////////////////////////////////////////////
/////////////////////////////////////////////
//LJ real

  // Get the first order approximation (division by 2^numiter)
  m_Divider->SetInput(inputPtr);
  m_Divider->GraftOutput( this->GetOutput() );
  m_Divider->SetConstant( static_cast<InputPixelRealValueType>(1<<numiter) );
  m_Divider->Update();

  // Region passing stuff
  this->GraftOutput( m_Divider->GetOutput() );
  this->GetOutput()->Modified();

  if (m_LogJacobianDeterminantComputation)
    {
    m_LogJacobianDetImage=Divergence<InputImageType,ScalarImageType>(m_Divider->GetOutput());
    m_ScalarWarper->SetOutputOrigin(this->GetOutput()->GetOrigin());
    m_ScalarWarper->SetOutputSpacing(this->GetOutput()->GetSpacing());
    m_ScalarWarper->SetOutputDirection(this->GetOutput()->GetDirection());
    }

  /** Initialize the warper for the iterative composition of the vector field **/
/*  m_VectorWarper->SetOutputOrigin(inputPtr->GetOrigin());
  m_VectorWarper->SetOutputSpacing(inputPtr->GetSpacing());
  m_VectorWarper->SetOutputDirection(inputPtr->GetDirection());

  /** Initialize the update for the exponential computation **/
/*  InputImagePointer ActualVectorField=m_Divider->GetOutput();

  /** First step of the exponential computation  **/
/*  this->GraftOutput( m_Divider->GetOutput() );
  this->GetOutput()->Modified();



  /**  Define the number of iteration based on the iterative scheme: scaling and squaring: N, small steps: 2^N **/
/*  if (m_IterativeScheme==FORWARD_EULER)
    numiter=static_cast<InputPixelRealValueType>(1<<numiter)-1;

  for( unsigned int i=0; i<numiter; i++ )
    {
    /** Compute the Log-Jacobian determinant inside whitin the exponential computation **/
/*    if (m_LogJacobianDeterminantComputation)
      {
      if (i==0)  /**  The initial approximation for the Jacobian is the divergence **/
 /*       m_ScalarWarper->SetInput(m_LogJacobianDetImage);
      else
        m_ScalarWarper->SetInput(m_ScalarAdder->GetOutput());

      m_ScalarWarper->SetDeformationField(ActualVectorField);
      m_ScalarWarper->UpdateLargestPossibleRegion();

      m_ScalarAdder->SetInput1(m_ScalarWarper->GetOutput());

      if (i==0||m_IterativeScheme==FORWARD_EULER)
        m_ScalarAdder->SetInput2(m_LogJacobianDetImage);
      else
        m_ScalarAdder->SetInput2(m_ScalarAdder->GetOutput());

      m_ScalarAdder->Update();
      }

    /** Compute the exponential by iterative composition **/

/*    m_VectorWarper->SetInput(this->GetOutput());

    m_VectorWarper->SetDeformationField(ActualVectorField);

    m_VectorWarper->GetOutput()->SetRequestedRegion(this->GetOutput()->GetRequestedRegion() );

    m_VectorWarper->Update();

    OutputImagePointer warpedIm = m_VectorWarper->GetOutput();
    warpedIm->DisconnectPipeline();

    m_VectorAdder->SetInput1(ActualVectorField);

    m_VectorAdder->SetInput2(warpedIm);
    m_VectorAdder->GetOutput()->SetRequestedRegion(
    this->GetOutput()->GetRequestedRegion() );

    m_VectorAdder->Update();

    // Region passing stuff
    this->GraftOutput( m_VectorAdder->GetOutput() );

    // Make a call to modified. This seems only necessary for
    // a non-inplace adder but it doesn't hurt anyhow.
    this->GetOutput()->Modified();

    if (m_IterativeScheme==SCALING_AND_SQUARING)
      ActualVectorField=this->GetOutput();

    }

  if (m_LogJacobianDeterminantComputation)
    m_LogJacobianDetImage=m_ScalarAdder->GetOutput();

  }


} // end namespace itk
*/


/*  if (param.NumericalScheme==1)
  {
   for (int i=0;i<numiter;++i)
    {
     WarpImg->SetInput(Image);
     VectorWarper->SetInput(Vect1);

     if (i==0)
     {
      WarpImg->SetDeformationField(Vect1);
     }
     else
     {
      WarpImg->SetDeformationField(VectorAdder->GetOutput());
     }

     WarpImg->SetOutputOrigin(Vect1->GetOrigin());
     WarpImg->SetOutputSpacing(Vect1->GetSpacing());
     WarpImg->SetOutputDirection(Vect1->GetDirection());
     WarpImg->UpdateLargestPossibleRegion();

     Add->SetInput1(WarpImg->GetOutput());
     if (i==0)
     {
      Add->SetInput2(Image);
     }
     else
     {
      Add->SetInput2(Add->GetOutput());
     }
     Add->Update();

     if (i==0)
     {
      VectorWarper->SetDeformationField(Vect1);
     }
     else
     {
      VectorWarper->SetDeformationField(VectorAdder->GetOutput());
     }
     VectorWarper->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion());
     VectorWarper->Update();
     warpedIm = VectorWarper->GetOutput();

     VectorAdder->SetInput1(Vect1);
     VectorAdder->SetInput2(VectorWarper->GetOutput());
     VectorAdder->GetOutput()->SetRequestedRegion(Vect1->GetRequestedRegion() );
     VectorAdder->Update();
    }
  }
  else
  {*/
